from .default import DefaultExecutor
from ..error import ExecutorFunctionNotExistError
from typing import Any, List


class JumperExecutor(DefaultExecutor):
    jump_functions = ["JUMP"]
    line = 0
    # Executor version - it's set on the lineup version
    _version = None

    def jump(self, line: int):
        self.line = line
        self.logger.debug(
            f"Jump to: {self.line} (execute index {self.line + 1} or line " +
            f"{self.line + 2})")

    def execute_jump(self, line: List[str]):
        if line[0] == "JUMP":
            if (len(line) > 2):
                if line[2] == "FROM":
                    self.jump(self.line + int(line[1]))
            else:
                self.jump(int(line[1]) - 2)
            return None
        raise ExecutorFunctionNotExistError(
            f"'{line[0]}' not exist in '{self}'")

    def execute_line(self, line: List[str]):
        if line[0] in self.jump_functions:
            return self.execute_jump(line)
        return super().execute_line(line)

    def execute(self, script: List[List[str]]) -> Any:
        self.stop = False
        result = None
        self.line = -1
        while self.line < len(script) - 1:
            if self.stop:
                break
            self.line += 1
            result = self.execute_line(script[self.line])
        return result



from .error import DecodeLineStringError
from typing import List
import regex as re


class LineDecoder:
    """
    Decode a line string to an instruction list
    """

    def decode(self, line: str) -> List[str] | None:
        """
        Decode a line string to an instruction list
        """
        line = line.strip()
        if line.startswith("#"):
            return None
        # delete comments on the line
        first_comment_block = re.search(r"[^\\]#", line)
        if first_comment_block:
            line = line[:first_comment_block.start()]
        line = line.replace("\\#", "#")  # replace escaped comments
        return self._split_line(line)

    def _split_line(self, line: str) -> List[str]:
        lines = self.format_quotes(line)
        result = []
        # odd is inside quotes, even is outside
        for i, line in enumerate(lines):
            if i % 2 == 1:
                # inside quotes, the line has to be kept as is
                result.append(line)
            elif line.strip() != "":
                # outside quotes, split by space for get the instructions
                result.extend(line.strip().split(" "))
        return result

    def format_quotes(self, line: str) -> List[str]:
        """
        Format the quotes in the line

        Split the line by quotes, keep the backslahed quotes
        Return a list of odd is inside quotes, even is outside the quotes
        """
        lines = line.split('"')
        result = []
        current_line = ""
        for i in lines:
            current_line += i
            if i.endswith("\\"):
                current_line += '"'
            else:
                result.append(current_line)
                current_line = ""
        if current_line:
            raise DecodeLineStringError(f"'{line}' is not valid line string")
        return result



from ..language_object import CoreObjectInterface, LanguageObjectInterface
from ..error import LineupError
from typing import Dict, Any, List


class VariableNotExistError(LineupError):
    pass


class DeleteDefaultVariableError(LineupError):
    pass


class VariableObject(CoreObjectInterface):
    variables: Dict[str, Any]
    default_variables: List[str]
    version = None

    def __init__(self, variables: Dict[str, Any] = {}) -> None:
        self.variables = variables
        self.default_variables = list(variables.keys())
        self.functions = {
            "EXIT": self._exit,
            "VAR": self._variable,
            "EXEC": self._execute_in_variables,
            "GET": self._get,
        }

    def close(self):
        for key in self.variables.keys():
            if type(self.variables[key]) is LanguageObjectInterface:
                self.variables[key].close()

    def reset(self):
        variables_delete = [key for key in self.variables.keys()
                            if key not in self.default_variables]
        for key in variables_delete:
            if type(self.variables[key]) is LanguageObjectInterface:
                self.variables[key].close()
            del self.variables[key]
        for key in self.variables.keys():
            if type(self.variables[key]) is LanguageObjectInterface:
                self.variables[key].reset()

    def _get(self, name: str):
        if name in self.variables:
            return self.variables[name]
        msg = f"'{name}' not exist in '{self}'"
        self.logger.error(msg)
        raise VariableNotExistError(msg)

    def _set(self, name: str, value):
        self.variables[name] = value

    def _delete(self, name: str):
        if name in self.default_variables:
            msg = f"'{name}' is default variable in '{self}'"
            self.logger.error(msg)
            raise DeleteDefaultVariableError(msg)
        if name in self.variables:
            del self.variables[name]

    def _execute_in_variables(self, variables, function_name: str, *args):
        if variables in self.variables:
            return self.variables[variables].execute(function_name, *args)
        return None

    def _execute_from_executor(self, line: List[str]):
        return self.executor.execute_line(line)

    def _exit(self, *args):
        self.executor.stop = True
        return self._execute_from_executor(args)

    def _variable(self, name: str, command: str, *args):
        if command == "USE":
            self._set(name, self._execute_in_variables(args[0], *args[1:]))
        elif command == "COPY":
            self._set(name, self._get(args[0]))
        elif command == "UNSET":
            self._delete(name)
        elif command == "GET":
            return self._get(name)
        elif command == "SET":
            self._set(name, args[0])
        elif command == "EXEC":
            return self._set(name, self._execute_from_executor(args))
        return None



from ..language_object import CoreObjectInterface
from ..error import LineupError
from typing import List, Tuple


class ConditionsJumpObject(CoreObjectInterface):
    last_result = None
    version = None

    def __init__(self):
        self.functions = {
            "IF": self._if,
            "NOTIF": self._notif,
            "ELSE": self._else,
            "EQUAL": self._equal,
        }

    def _get_two_command(self, jump_size: str, from_or_first_arg: str,
                         *args) -> Tuple[List[str], List[str]]:
        command = args
        jump_command = ["JUMP", jump_size]
        if from_or_first_arg == "FROM":
            jump_command.append("FROM")
        else:
            command.insert(0, from_or_first_arg)
        return [jump_command, command]

    def _if(self, jump_size: str, from_or_first_arg: str, *args):
        jump_command, command = self._get_two_command(jump_size,
                                                      from_or_first_arg,
                                                      *args)
        result = False
        try:
            result = self.executor.execute_line(command)
        except LineupError:
            pass
        if result:
            self.last_result = True
            self.executor.execute_line(jump_command)
        self.last_result = False

    def _notif(self, jump_size: str, from_or_first_arg: str, *args):
        jump_command, command = self._get_two_command(jump_size,
                                                      from_or_first_arg,
                                                      *args)
        result = False
        try:
            result = self.executor.execute_line(command)
        except LineupError:
            pass
        if not result:
            self.last_result = True
            self.executor.execute_line(jump_command)
        self.last_result = False

    def _else(self, *args):
        if self.last_result:
            raise LineupError("ELSE without IF")
        self.last_result = False
        self.executor.execute_line(["JUMP", *args])

    def _equal(self, *args):
        first = self.executor.execute_line(["VAR", args[0], "GET"])
        second = self.executor.execute_line(["VAR", args[1], "GET"])
        if first == second:
            self.last_result = True
        else:
            self.last_result = False
        return self.last_result



from setuptools import setup
from pathlib import Path


setup(
    name="lineup-lang",
    version=Path(__file__).parent.joinpath("lineup_lang/VERSION").read_text(),
    description="Pseudo language interpreter for Python",
    long_description=Path(__file__).parent.joinpath("README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://github.com/alex-bouget/Lineup",
    author="alex-bouget",
    packages=["lineup_lang", "lineup_lang.core", "lineup_lang.executor"],
    package_data={"lineup_lang": ["VERSION"]},
    install_requires=[
        "regex",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Interpreters",
    ]
)
